/*====================================+=====================================+
! File CConfig.h                      ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#pragma once
/*-------------------------------------------------------------------------*/
#include <wx/crt.h>
#include <wx/config.h>
#include <wx/gdicmn.h>
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
template < class T >
class CConfigBaseElement
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CConfigBaseElement()
      {  ; }
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      wxString m_s_key       ;
      bool     m_boo_to_save = false ; // Val changed => has to be saved
      T        m_val         ;
      T        m_def_val     ; // Default value
   /*----------------------------------------------------------------------*/
   public :
      /*--------------------------------------------------------------------+
      ! Returning values by reference. Could think it might create overhead !
      ! for int and bool but it's not.                                      !
      +--------------------------------------------------------------------*/
      const T &get() const { return( m_val ) ; }
      const T &get_def() const { return( m_def_val ) ; }
      /*-------------------------------------------------------------------*/
      void set( const T &val )
      {  if( m_val != val ) { m_boo_to_save = true ; m_val = val ; } }
      bool get_to_save() const { return( m_boo_to_save ) ; }
      void set_to_save( bool boo_val ) { m_boo_to_save = boo_val ; }
      /*-------------------------------------------------------------------*/
      void save() { if( m_boo_to_save ) { write_val() ; } }
      /*-------------------------------------------------------------------*/
      const wxString &get_key() const { return( m_s_key ) ; }
      void set_key( const wxString &s_key ) { m_s_key = s_key ; }
      /*-------------------------------------------------------------------*/
      void init( const wxString &s_key, const T &def_val = T() )
      {  set_key( s_key ) ; m_def_val = def_val ; read_val( m_def_val ) ; }
      /*--------------------------------------------------------------------+
      ! "Virtual" to call the right function from "save" etc.               !
      +--------------------------------------------------------------------*/
      virtual void write_val() = 0 ;
      virtual void read_val( const T &def_val ) = 0 ;
      /*-------------------------------------------------------------------*/
      void del()
      {
         /*----------------------------------------------------------------*/
         wxConfigBase *p_conf = wxConfigBase::Get() ;
         if( p_conf == nullptr ) { return ; }
         /*----------------------------------------------------------------*/
         p_conf->DeleteEntry( m_s_key ) ;
         m_boo_to_save = false ;
         /*----------------------------------------------------------------*/
      }
   /*----------------------------------------------------------------------*/
} ;

/*--------------------------------------------------------------------------+
! - Why defining the class CConfigBase and not define the virtual           !
!   functions directly in CCongigBaseElement ?                              !
! * This seems related to a problem between gcc and the "template"          !
!                                                                           !
! - Why using "this->" in CConfigBase for variables defined protected in    !
!   CConfigBaseElement ?                                                    !
! * This seems related to a problem between gcc and the "template" class    !
!   derivation                                                              !
+--------------------------------------------------------------------------*/
template < class T >
class CConfigBase : public CConfigBaseElement< T >
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CConfigBase< T >() : CConfigBaseElement< T >()
      {  ; }
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      void write_val()
      {
         /*----------------------------------------------------------------*/
         wxConfigBase *p_conf = wxConfigBase::Get() ;
         if( p_conf == nullptr ) { return ; }
         /*----------------------------------------------------------------*/
         p_conf->Write( this->m_s_key, this->m_val ) ;
         this->m_boo_to_save = false ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      void read_val( const T &def_val )
      {
         /*----------------------------------------------------------------*/
         wxConfigBase *p_conf = wxConfigBase::Get() ;
         if( p_conf == nullptr ) { return ; }
         /*--( Get file data )---------------------------------------------*/
         p_conf->Read( this->m_s_key, &this->m_val, def_val ) ;
         this->m_boo_to_save = false ;
         /*----------------------------------------------------------------*/
      }
   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
using CConfigInt    = CConfigBase< int > ;
using CConfigDouble = CConfigBase< double > ;

/*-------------------------------------------------------------------------*/
class CConfigBool : public CConfigBase< bool >
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CConfigBool() : CConfigBase< bool >()
      {  ; }
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      void toggle()
      {  set( !get() ) ; }

   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
class CConfigString : public CConfigBase< wxString >
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CConfigString() : CConfigBase< wxString >()
      {  ; }
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      bool empty() const { return( m_val.empty() ) ; }
      mtf::wxString_cit begin() { return( m_val.begin() ) ; }
      mtf::wxString_cit end() { return( m_val.end() ) ; }
      void clear()
      {  if( !m_val.empty() )
         {  m_val.clear() ; set_to_save( true ) ; }
      }
      void append( const wxString &s )
      {  m_val += s ; set_to_save( true ) ; }
      /*--( Init for strings that can't be empty )-------------------------*/
      void init_mandatory( const wxString &s_key,
                           const wxString &def_val = wxString()
                         )
      {  init( s_key, def_val ) ;
         if( empty() ) { set( get_def() ) ; }
      }

   /*----------------------------------------------------------------------*/
} ;

/*--------------------------------------------------------------------------+
! I could have defined the wxRect specialized functions but it doesn't seem !
! to work perfectly under gcc (or maybe it's me!)                           !
+--------------------------------------------------------------------------*/
class CConfigRect : public CConfigBaseElement< wxRect >
{
   /*----------------------------------------------------------------------*/
   public :
      /*-------------------------------------------------------------------*/
      CConfigRect() : CConfigBaseElement< wxRect >()
      {  ; }
   /*----------------------------------------------------------------------*/
   public :
      /*--( For this type, default value will be "specific" )--------------*/
      void init( const wxString &s_key,
                 const wxRect &def_val
                            = wxRect( wxDefaultPosition, wxDefaultSize )
               )
      {  CConfigBaseElement< wxRect >::init( s_key, def_val ) ; }
      /*-------------------------------------------------------------------*/
      void write_val()
      {
         /*----------------------------------------------------------------*/
         wxConfigBase *p_conf = wxConfigBase::Get() ;
         if( p_conf == nullptr ) { return ; }
         /*----------------------------------------------------------------*/
         p_conf->Write( m_s_key,
                        wxString::Format( "%d %d %d %d",
                                          m_val.x, m_val.y,
                                          m_val.width, m_val.height
                                        )
                      ) ;
         /*----------------------------------------------------------------*/
         m_boo_to_save = false ;
         /*----------------------------------------------------------------*/
      }

      /*-------------------------------------------------------------------*/
      void read_val( const wxRect &def_val )
      {
         /*----------------------------------------------------------------*/
         wxConfigBase *p_conf = wxConfigBase::Get() ;
         if( p_conf == nullptr ) { return ; }
         /*--( Get file data )---------------------------------------------*/
         wxString s_val ;
         if( !p_conf->Read( m_s_key, &s_val ) )
         {  m_val = def_val ; }
         else
         {  wxSscanf( s_val, "%d %d %d %d",
                      &m_val.x, &m_val.y, &m_val.width, &m_val.height
                    ) ;
         }
         /*----------------------------------------------------------------*/
         m_boo_to_save = false ;
         /*----------------------------------------------------------------*/
      }
   /*----------------------------------------------------------------------*/
} ;

/*-------------------------------------------------------------------------*/
void CConfig_open( const wxString &s_file, const int i_version ) ;

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                        End of file CConfig.h                              !
+==========================================================================*/
