/*====================================+=====================================+
! File CApp.cpp                       ! Copyright (C) 2016-2018 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of matelefree.                                          !
! matelefree is free software: you can redistribute it and/or modify it     !
! under the terms of the GNU General Public License as published by the     !
! Free Software Foundation, either version 3 of the License, or any later   !
! version.                                                                  !
! matelefree is distributed in the hope that it will be useful, but WITHOUT !
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or     !
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for !
! more details.                                                             !
! You should have received a copy of the GNU General Public License along   !
! with matelefree. If not, see <http://www.gnu.org/licenses/>.              !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <memory>
#include <wx/imaglist.h>
#include <wx/stdpaths.h>
#include <wx/filename.h>
#include <wx/dir.h>
#include <wx/wfstream.h>
#include <wx/txtstrm.h>
#include <wx/url.h>
#include <wx/intl.h>
#include <wx/fs_zip.h>
#include <wx/regex.h>
#include <wx/tokenzr.h>
#include <wx/textfile.h>
#include <wx/sstream.h>
#include <curl/curl.h>
/*-------------------------------------------------------------------------*/
#include "matelefree_info.h"
#include "CApp.h"
#include "CFrame.h"
#include "CConfig.h"
#include "CDownload.h"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
CApp::~CApp()
{
   /*----------------------------------------------------------------------*/
   for( auto it : m_map_channel ) { delete it.second ; }
   /*----------------------------------------------------------------------*/
   delete m_il_logo ;
   /*----------------------------------------------------------------------*/
   curl_global_cleanup() ;
   /*----------------------------------------------------------------------*/
   if( !m_s_program_to_start_on_exit.empty() )
   {  wxExecute( m_s_program_to_start_on_exit ) ; }
   /*----------------------------------------------------------------------*/
}

/*--( Return 0 if the program has not to exit )----------------------------*/
int CApp::actions_for_install_process()
{
   /*----------------------------------------------------------------------*/
   wxLogNull no_log ; // Suppress useless warnings

   /*-----------------------------------------------------------------------+
   ! The "autoinstall" process has to replace the "matelefree" exec file,   !
   ! the one currently being executed. The problem is that an executed      !
   ! program file is "locked".                                              !
   ! This modification is done in steps.                                    !
   !                                                                        !
   ! The current state of the installation is indicated by the executable   !
   ! base name :                                                            !
   ! - matelefree : Standard execution                                      !
   !   Its tasks are : - delete "matelefree_new" if present                 !
   !                   - Assume expected other tasks :)))                   !
   !                                                                        !
   ! - matelefree_new : After an automatic installation the new executable  !
   !   has been started by "matelefree"                                     !
   !   Its tasks are : - copy itself to the std "matelefree" executable     !
   !                   - start the standard "matelefree" executable         !
   +-----------------------------------------------------------------------*/
   wxFileName fn_exe( wxStandardPaths::Get().GetExecutablePath() ) ;

   /*----------------------------------------------------------------------*/
   if( fn_exe.GetName().EndsWith( "_new" ) )
   {
      /*-------------------------------------------------------------------*/
      wxFileName fn_std( fn_exe ) ;
      fn_std.SetName( fn_std.GetName().BeforeFirst( '_' ) ) ;
      /*--( Whatever the result is, a restart is done on std executable )--*/
      set_program_to_start_on_exit( fn_std.GetFullPath() ) ;
      /*--------------------------------------------------------------------+
      ! The standard exe can still been locked.                             !
      ! This loop should not be infinite or there is a bigger problem       !
      +--------------------------------------------------------------------*/
      while( !wxCopyFile( fn_exe.GetFullPath(), fn_std.GetFullPath() ) )
      {  wxMilliSleep( 1000 ) ; }
      /*-------------------------------------------------------------------*/
      return( 1 ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Remove "new" if present and continue )----------------------------*/
   wxFileName fn_new( fn_exe ) ;
   fn_new.SetName( fn_new.GetName() + "_new" ) ;
   /*-----------------------------------------------------------------------+
   ! The new exe can still been locked.                                     !
   ! This loop should not be infinite or there is a bigger problem          !
   +-----------------------------------------------------------------------*/
   while( fn_new.FileExists() && !wxRemoveFile( fn_new.GetFullPath() ) )
   {  wxMilliSleep( 1000 ) ; }
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CApp::OnInit()
{
   /*----------------------------------------------------------------------*/
   curl_global_init( CURL_GLOBAL_ALL ) ;
   /*----------------------------------------------------------------------*/
   m_fn_executable = wxStandardPaths::Get().GetExecutablePath() ;

   /*--( Quit if the install process started another version )-------------*/
   if( actions_for_install_process() != 0 ) { return( false ) ; }

   /*----------------------------------------------------------------------*/
   m_locale.Init( wxLANGUAGE_DEFAULT, wxLOCALE_DONT_LOAD_DEFAULT ) ;
   /*--( During development locale dir is located upper ... )--------------*/
   m_locale.AddCatalogLookupPathPrefix(   get_exe_path()
                                        #ifdef __WXDEBUG__
                                        + "/../../.."
                                        #endif // __WXDEBUG__
                                        + "/locale"
                                      ) ;
   /*--( The command line analyse and parsing is done via this call )------*/
   if( !wxApp::OnInit() ) { return( false ) ; }
   /*--( Done avec "ini" loading )-----------------------------------------*/
   init_fn_vlc() ;
   /*--( By default, the application name is set to argv[0] )--------------*/
   SetAppName( MTF_NAME ) ;
   /*--( Init stream class static elements )-------------------------------*/
   str_stream::st_init() ;
   /*--( Image list & archive )--------------------------------------------*/
   wxImage::AddHandler( new wxPNGHandler ) ;
   init_imagelist() ;
   wxFileSystem::AddHandler( new wxZipFSHandler ) ;

   /*--( Init in two steps to set "m_frame" and use it during init )-------*/
   m_frame = new CFrame() ;
   m_frame->init() ;
   /*----------------------------------------------------------------------*/
   return( true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::OnInitCmdLine( wxCmdLineParser &parser )
{
   /*----------------------------------------------------------------------*/
   static const wxCmdLineEntryDesc st_co_tb_cmd_line_desc[] =
   {
      { wxCMD_LINE_SWITCH, "h", "help",
        "show this help message",
        wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP
      },
      { wxCMD_LINE_SWITCH, "v", "version",
        "print version information"
      },
      { wxCMD_LINE_SWITCH, "F", "forcefreebox",
        "force freebox access even if not accessible"
      },
      { wxCMD_LINE_SWITCH, "M", "forceminimalest",
        "force smallest minimal display (no menu, no caption) "
        "when minimal display is activated. "
        "It may not work depending on the system."
      },
      { wxCMD_LINE_OPTION, "l", "language",
        "language (en, en_US, fr_FR ...)"
      },
      { wxCMD_LINE_SWITCH, "N", "novlcinitstr",
        "do not use any \"VLC init strings\""
      },
      { wxCMD_LINE_NONE }
   } ;

   /*----------------------------------------------------------------------*/
   parser.SetLogo( MTF_DISPLAY_NAME_ASCII " " MTF_VERSION "\n"
                   MTF_COPYRIGHT_ASCII " (" MTF_HOST ")\n"
                   MTF_LICENSE_ASCII "\n"
                 ) ;
   parser.SetDesc( st_co_tb_cmd_line_desc ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CApp::OnCmdLineParsed( wxCmdLineParser &parser )
{
   /*--( Version information => quit )-------------------------------------*/
   if( parser.Found( "v" ) )
   {  wxMessageOutput::Get()->Printf( "%s %s", MTF_DISPLAY_NAME, MTF_VERSION
                                    ) ;
      return( false ) ;
   }

   /*----------------------------------------------------------------------*/
   const wxChar co_c_path_sep = wxFileName::GetPathSeparator() ;
   wxFileName fn_media_path = wxFileName::DirName( get_exe_path() ) ;

   /*--( Debug executables are in "mtf/comp/bin/debug" )-------------------*/
   #ifdef __WXDEBUG__
      for( int i_num = 0 ; i_num < 4 ; ++i_num )
      {  fn_media_path.RemoveLastDir() ; }
   #endif // DEBUG
   fn_media_path.AppendDir( "data" ) ;
   /*----------------------------------------------------------------------*/
   m_s_media_path    = fn_media_path.GetFullPath() ;
   m_s_media_fb_path = m_s_media_path + "freebox" + co_c_path_sep ;
   /*----------------------------------------------------------------------*/
   m_s_tree_fb_path = "Freebox" ;
   m_s_tree_fbtv_path =   m_s_tree_fb_path
                        + '/' + wxGetTranslation( get_pk_label( PK_FBTV ) ) ;
   /*-----------------------------------------------------------------------+
   ! Under MSW the config file will be in the executable's directory        !
   ! Under GTK (and others) the default is used: ~/.appname                 !
   +-----------------------------------------------------------------------*/
   wxString s_config_filename ;
   /*----------------------------------------------------------------------*/
   #ifdef __WXMSW__
      /*-------------------------------------------------------------------*/
      s_config_filename = MTF_NAME ".ini" ;
      /*-------------------------------------------------------------------*/
      wxFileName fn_config_file( s_config_filename ) ;
      if( fn_config_file.MakeAbsolute( get_exe_path() ) )
      {  s_config_filename = fn_config_file.GetFullPath() ; }
      /*-------------------------------------------------------------------*/
   #elif defined( __WXGTK__ )
        s_config_filename
      = wxFileConfig::GetLocalFile( MTF_NAME ).GetFullPath() ;
   #else
      #error Unknown platform to: determine the config file full name
   #endif // __WXGTK__
   /*-----------------------------------------------------------------------+
   ! Initialization of the configuration file.                              !
   ! If it doesn't exist, it means that the program has never been started  !
   +-----------------------------------------------------------------------*/
   m_boo_first_start_ever = !wxFileExists( s_config_filename ) ;
   /*--( The current config version of this version is set here )----------*/
   CConfig_open( s_config_filename, 2 ) ;
   init_config_info() ;

   /*----------------------------------------------------------------------*/
   m_boo_forcefreebox     = parser.Found( "F" ) ;
   m_boo_force_minimalest = parser.Found( "M" ) ;
   /*----------------------------------------------------------------------*/
   wxString s_canonical_name ;
   /*--( Language specification precedence: 1) command line, 2) system )---*/
   if( parser.Found( "l", &s_canonical_name ) )
   {  M_s_lang.set( s_canonical_name ) ; }
   /*----------------------------------------------------------------------*/
   init_translations( M_s_lang.get() ) ;
   /*----------------------------------------------------------------------*/
   m_boo_use_vlc_init_str = !parser.Found( "N" ) ;
   /*----------------------------------------------------------------------*/
   return( true ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_fn_vlc()
{
   /*----------------------------------------------------------------------*/
   wxString s_path = M_s_vlc_path.get() ;
   /*--( We will just have to find out the vlc program path )--------------*/
   #ifdef __WXMSW__
      /*-------------------------------------------------------------------*/
      m_fn_vlc.SetFullName( "vlc.exe" ) ;
      /*-------------------------------------------------------------------*/
      if( s_path.empty() )
      {
         /*----------------------------------------------------------------*/
         wxLogNull no_log ; // Suppress useless warnings
         /*----------------------------------------------------------------*/
         if( !wxRegKey( wxRegKey::HKLM, R"=(Software\VideoLAN\VLC)="
                      ).QueryValue( "InstallDir", s_path )
           )
         {  wxRegKey( wxRegKey::HKLM,
                      R"=(Software\Wow6432Node\VideoLAN\VLC)="
                    ).QueryValue( "InstallDir", s_path ) ;
         }

         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   #else
      m_fn_vlc.SetFullName( "vlc" ) ;
   #endif // __WXMSW__
   /*----------------------------------------------------------------------*/
   m_fn_vlc.SetPath( s_path ) ;
   /*--( Finally try to find VLC via the "PATH" variable )-----------------*/
   if( !m_fn_vlc.Exists() )
   {  /*-------------------------------------------------------------------*/
      wxPathList path_list ;
      path_list.AddEnvList( "PATH" ) ;
      s_path = path_list.FindAbsoluteValidPath( m_fn_vlc.GetFullName() ) ;
      if( !s_path.empty() ) { m_fn_vlc = s_path ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_imagelist()
{
   /*----------------------------------------------------------------------*/
   const int co_i_x = 70 ;
   const int co_i_y = 54 ;
   /*----------------------------------------------------------------------*/
   m_il_logo = new wxImageList( co_i_x, co_i_y, false ) ;
   /*-----------------------------------------------------------------------+
   ! This "dummy image" will give data to wxGTK wxGenericListCtrl to        !
   ! compute the line height. Else it is computed based only on text size.  !
   +-----------------------------------------------------------------------*/
   m_il_logo->Add( wxImage( co_i_x, co_i_y, true ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_translations( const wxString &s_canonical_name )
{
   /*--( Default is the system language ... )------------------------------*/
   if( !s_canonical_name.empty() )
   {
      /*--( This will only change the texts, not the locales )-------------*/
      wxTranslations *p_translations = new wxTranslations() ;
      p_translations->SetLanguage( s_canonical_name ) ;
      wxTranslations::Set( p_translations ) ;
      /*-------------------------------------------------------------------*/
   }
   /*--( The standard catalog is added for "Copy/Paste/Delete/Cancel ..." )*/
   m_locale.AddCatalog( "wxstd" ) ;
   /*----------------------------------------------------------------------*/
   m_locale.AddCatalog( MTF_NAME ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannel *CApp::find_channel( const wxString &s_name ) const
{
   /*----------------------------------------------------------------------*/
   auto it = m_map_channel.find( s_name ) ;
   /*----------------------------------------------------------------------*/
   return( it != m_map_channel.end() ? it->second : nullptr ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CChannel *CApp::find_channel( int i_num ) const
{
   /*----------------------------------------------------------------------*/
   auto it = m_map_channel_num.find( i_num ) ;
   /*----------------------------------------------------------------------*/
   return( it != m_map_channel_num.end() ? it->second : nullptr ) ;
   /*----------------------------------------------------------------------*/
}

/*--( To be used before the reset of config file )-------------------------*/
void CApp::init_config_info_to_keep()
{
   /*--( The entries we want to keep )-------------------------------------*/
   M_s_bopl_pl_user_list.init( "/current/bopl_pl_user_list" ) ;
   M_s_lang.init( "/current/lang" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_config_info()
{
   /*----------------------------------------------------------------------*/
   init_config_info_to_keep() ;
   /*----------------------------------------------------------------------*/
   M_s_main_perspective.init( "/current/main_perspective" ) ;
   M_s_bopl_expanded.init( "/current/bopl_expanded" ) ;
   M_main_rec.init( "/current/main_rec" ) ;
   M_main_mini_rec.init( "/current/main_mini_rec" ) ;
   M_boo_mini_display.init( "/current/mini_display", false ) ;
   M_boo_mini_display_caption.init( "/current/mini_display_caption", true ) ;
   M_boo_stay_on_top.init( "/current/stay_on_top", false ) ;
   M_s_bopl_sel_name.init( "/current/bopl_sel_name" ) ;
   /*----------------------------------------------------------------------*/
   M_boo_freebox.init( "/current/freebox", true ) ;
   M_boo_mute.init( "/current/mute", false ) ;
   M_i_max_nb_media.init( "/current/max_nb_media", 2 ) ;
   M_boo_new_media_active.init( "/current/new_media_active", true ) ;
   M_i_check_media_delay.init( "/current/check_media_delay", 2000 ) ;
   M_i_volume.init( "/current/volume", 100 ) ;
   /*----------------------------------------------------------------------*/
   M_prefs_rec.init( "/current/prefs_rec" ) ;
   M_s_vlc_path.init( "/current/vlc_path" ) ;
   M_s_snapshot_path.init( "/current/snapshot_path" ) ;
   M_boo_mute_deact.init( "/current/mute_deact", false ) ;
   M_boo_startup_play.init( "/current/startup_play", true ) ;
   /*--( This client ID is "matelefree's" one )----------------------------*/
   M_s_twitch_client_id.init( "/current/twitch_client_id",
                              "asvncavqk0no3newx9ukaz4i230vqz"
                            ) ;
   M_s_proxy_server.init( "/current/proxy_server" ) ;
   M_s_startup_play_list.init( "/current/startup_play_list" ) ;
   M_s_startup_play_active.init( "/current/startup_play_active" ) ;
   M_s_play_hist.init( "/current/play_hist" ) ;
   M_s_vlc_init_dm.init( "/current/vlc_init_dm" ) ;
   M_s_vlc_init_tw.init( "/current/vlc_init_tw" ) ;
   M_s_vlc_init_yt.init( "/current/vlc_init_yt" ) ;
   M_s_vlc_init_us.init( "/current/vlc_init_us") ;
   M_s_vlc_init_pe.init( "/current/vlc_init_pe") ;
   M_s_vlc_init_mx.init( "/current/vlc_init_mx") ;
   M_s_vlc_init_other.init( "/current/vlc_init_other" ) ;
   M_s_vlc_init_all.init( "/current/vlc_init_all" ) ;
   M_s_deinterlace.init( "/current/deinterlace" ) ;
   M_i_chkupd_freq.init( "/current/chkupd_freq", CUF_WEEK ) ;
   M_s_chkupd_last.init( "/current/chkupd_last" ) ;
   M_s_def_quality.init( "/current/def_quality" ) ;
   M_i_viewers_delay.init( "/current/viewers_delay", 60 ) ;
   M_i_retry_delay.init( "/current/retry_delay", 60 ) ;
   /*----------------------------------------------------------------------*/
   M_progress_rec.init( "/current/progress_rec" ) ;
   M_lastversion_rec.init( "/current/lastversion_rec" ) ;
   M_newversion_rec.init( "/current/newversion_rec" ) ;
   /*----------------------------------------------------------------------*/
   M_plname_rec.init( "/current/plname_rec" ) ;
   M_chname_rec.init( "/current/chname_rec" ) ;
   M_addlive_rec.init( "/current/addlive_rec" ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::save_config_info()
{
   /*----------------------------------------------------------------------*/
   M_main_rec.save() ;
   M_main_mini_rec.save() ;
   M_boo_mini_display.save() ;
   M_boo_mini_display_caption.save() ;
   M_boo_stay_on_top.save() ;
   M_s_bopl_sel_name.save() ;
   /*----------------------------------------------------------------------*/
   M_boo_freebox.save() ;
   M_boo_mute.save() ;
   M_i_max_nb_media.save() ;
   M_boo_new_media_active.save() ;
   M_i_check_media_delay.save() ;
   M_i_volume.save() ;
   /*----------------------------------------------------------------------*/
   M_prefs_rec.save() ;
   M_s_lang.save() ;
   M_s_vlc_path.save() ;
   M_s_snapshot_path.save() ;
   M_boo_mute_deact.save() ;
   M_boo_startup_play.save() ;
   M_s_twitch_client_id.save() ;
   M_s_proxy_server.save() ;
   M_s_startup_play_list.save() ;
   M_s_startup_play_active.save() ;
   M_s_play_hist.save() ;
   M_s_vlc_init_dm.save() ;
   M_s_vlc_init_tw.save() ;
   M_s_vlc_init_yt.save() ;
   M_s_vlc_init_us.save() ;
   M_s_vlc_init_pe.save() ;
   M_s_vlc_init_mx.save() ;
   M_s_vlc_init_other.save() ;
   M_s_vlc_init_all.save() ;
   M_s_deinterlace.save() ;
   M_i_chkupd_freq.save() ;
   M_s_chkupd_last.save() ;
   M_s_def_quality.save() ;
   M_i_viewers_delay.save() ;
   M_i_retry_delay.save() ;
   /*----------------------------------------------------------------------*/
   M_progress_rec.save() ;
   M_lastversion_rec.save() ;
   M_newversion_rec.save() ;
   /*----------------------------------------------------------------------*/
   M_plname_rec.save() ;
   M_chname_rec.save() ;
   M_addlive_rec.save() ;
   /*----------------------------------------------------------------------*/
   save_channel_config_info() ;
   save_playlist_config_info() ;
   /*----------------------------------------------------------------------*/
   M_s_main_perspective.save() ;
   M_s_bopl_pl_user_list.save() ;
   M_s_bopl_expanded.save() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CApp::add_channel( CChannel *p_channel )
{
   /*----------------------------------------------------------------------*/
   if( p_channel == nullptr ) { return( -1 ) ; }
   /*----------------------------------------------------------------------*/
   if( p_channel->is_num_set() )
   {  m_map_channel_num.emplace( p_channel->get_num(), p_channel ) ; }
   /*----------------------------------------------------------------------*/
   m_map_channel.emplace( p_channel->get_name(), p_channel ) ;
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::add_channel_m3u( const wxString &s_pl, const wxString &s_name,
                            const wxString &s_url_m3u,
                            mtf::t_vec_string &vs_channel
                          )
{
   /*----------------------------------------------------------------------*/
   CChannelDynamicKnown *p_channel
                      = new CChannelDynamicKnown( s_pl, s_name, s_url_m3u ) ;
   add_channel( p_channel ) ;
   vs_channel.emplace_back( p_channel->get_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::add_channel( CChannelDynamicKnown::e_cdtk cdtk,
                        const wxString &s_pl,
                        const wxString &s_id, const wxString &s_name,
                        mtf::t_vec_string &vs_channel
                      )
{
   /*----------------------------------------------------------------------*/
   CChannelDynamicKnown *p_channel
                     = new CChannelDynamicKnown( cdtk, s_pl, s_id, s_name ) ;
   add_channel( p_channel ) ;
   vs_channel.emplace_back( p_channel->get_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::add_channel_dailymotion( const wxString &s_pl,
                                    const wxString &s_id,
                                    const wxString &s_name,
                                    mtf::t_vec_string &vs_channel
                                  )
{
   /*----------------------------------------------------------------------*/
   CChannelDailymotion *p_channel
                     = new CChannelDailymotion( s_pl, s_id, s_name ) ;
   add_channel( p_channel ) ;
   vs_channel.emplace_back( p_channel->get_name() ) ;
   /*----------------------------------------------------------------------*/
}

/*--( This should only by called for "live" channels )---------------------*/
int CApp::del_channel( const mtf::t_vec_string &vs_channel )
{
   /*----------------------------------------------------------------------*/
   int i_nb_deleted = 0 ;
   wxConfigBase *p_conf = wxConfigBase::Get() ;
   p_conf->SetPath( CChannel::m_st_co_p_c_cfg_path ) ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_name : vs_channel )
   {
      /*--( The channel should be present in the name map )----------------*/
      CChannel *p_channel = get_channel( s_name ) ;
      if( p_channel == nullptr ) { return( -1 ) ; }
      /*--( Remove the entry in the num map )------------------------------*/
      if( p_channel->is_num_set() )
      {  m_map_channel_num.erase( p_channel->get_num() ) ; }
      /*--( Delete config entry of this channel in CURRENT PATH )----------*/
      p_conf->DeleteEntry( s_name ) ;
      /*--( Finally delete main element )----------------------------------*/
      m_map_channel.erase( s_name ) ;
      /*-------------------------------------------------------------------*/
      ++i_nb_deleted ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( i_nb_deleted ) ;
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Currently thjs function should only by called for "live" channels         !
! Note that the renaming in the user and search playlists is done in the    !
! calling function.                                                         !
! The channel number map is not concerned by the renaming ...               !
+--------------------------------------------------------------------------*/
int CApp::rename_channel( const wxString &s_from_full,
                          const wxString &s_to_short
                        )
{
   /*----------------------------------------------------------------------*/
   CChannel *p_channel = get_channel( s_from_full ) ;
   /*--( The channel should be present in the name map )-------------------*/
   if( p_channel == nullptr ) { return( -1 ) ; }
   /*--( First set object data )-------------------------------------------*/
   p_channel->rename( s_to_short ) ;
   /*--( Rename its entry in the config file )-----------------------------*/
   wxConfigBase *p_conf = wxConfigBase::Get() ;
   p_conf->SetPath( CChannel::m_st_co_p_c_cfg_path ) ;
   p_conf->RenameEntry( s_from_full, p_channel->get_name() ) ;
   /*--( Finally rename main entry index by recreating it )----------------*/
   m_map_channel.erase( s_from_full ) ;
   m_map_channel[ p_channel->get_name() ] = p_channel ;
   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CApp::load_channels( const wxString &s_origin,
                         const wxString &s_pl_name,
                         wxInputStream *p_is,
                         mtf::t_vec_string &vs_channel, // Append to it
                         wxString *p_s_label
                       )
{
   /*----------------------------------------------------------------------*/
   const e_pl_known  pl_known = find_pl_known( s_pl_name ) ;
   wxTextInputStream tis( *p_is, "", wxConvUTF8 ) ;
   wxString          s_extm3u_param ;
   wxString          s_line ;
   int               i_line_num = 0 ;

   /*----------------------------------------------------------------------*/
   if( p_s_label != nullptr ) { p_s_label->clear() ; }
   /*----------------------------------------------------------------------*/
   s_line = mtf::read_m3u_line( tis, i_line_num ) ;
   /*--( Check first line )------------------------------------------------*/
   if( s_line.StartsWith( "#EXTM3U", &s_extm3u_param ) )
   {
      /*--( Search for parameters in the first line )----------------------*/
      if( s_extm3u_param.StartsWith( ' ' ) )
      {
         /*--( Extract the tree label stored in a specific "parameter" )---*/
         if( p_s_label != nullptr )
         {  /*-------------------------------------------------------------*/
            mtf::match( s_extm3u_param, R"=(\mmtf-label\s*=\s*"([^"]*)")=",
                        p_s_label
                      ) ;
            /*-------------------------------------------------------------*/
         }
         /*--( Some m3u data have to be read from an URL ? )---------------*/
         wxString s_url ;
         if(    mtf::match( s_extm3u_param,
                            R"=(\mmtf-url\s*=\s*"([^"]*)")=", &s_url
                          ) > 1
             && !s_url.empty()
           )
         {  /*-------------------------------------------------------------*/
            if( m_boo_loading_channels_from_url )
            {  wxLogError( _( "\"%s\" line %d : A remote m3u file cannot "
                              "include another remote m3u file :\n%s"
                            ),
                           s_origin, i_line_num, s_url
                         ) ;
               return( -1 ) ;
            }
            /*-------------------------------------------------------------*/
            if( load_channels_from_url( s_url, s_pl_name, vs_channel ) != 0 )
            {  wxLogError( _( "\"%s\" line %d : "
                              "Error loading m3u file :\n%s"
                            ),
                           s_origin, i_line_num, s_url
                         ) ;
               return( -2 ) ;
            }
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
      s_line = mtf::read_m3u_line( tis, i_line_num ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Parsing a "title" line )------------------------------------------*/
   wxRegEx re_extinf( R"=(^#EXTINF: *-?\d+ *([^,]*)? *, *)="
                      R"=((-?\d*) *-? *(.*?)(\(([^\)]*?)\))?$)=",
                      wxRE_ADVANCED
                    ) ;
   /*----------------------------------------------------------------------*/
   while( !tis.GetInputStream().Eof() )
   {
      /*-------------------------------------------------------------------*/
      if( !s_line.StartsWith( "#EXTINF:" ) )
      {  wxLogError( _( "\"%s\" line %d : \"#EXTINF\" not found" ),
                     s_origin, i_line_num
                   ) ;
         return( -3 ) ;
      }
      /*--------------------------------------------------------------------+
      ! To be more consistent some modifications have to be made to         !
      ! "freebox" data.                                                     !
      ! By default some channel names available through TNT sees "HD" added !
      ! to them instead of adding it to the stream name (between "()").     !
      +--------------------------------------------------------------------*/
      if( s_line.EndsWith( " HD (TNT)", &s_line ) )
      {  s_line += " (TNT HD)" ; }

      /*-------------------------------------------------------------------*/
      if( !re_extinf.Matches( s_line ) )
      {  wxLogError( _( "\"%s\" line %d : Syntax error" ),
                     s_origin, i_line_num
                   ) ;
         return( -4 ) ;
      }

      /*--( Extract channel name )-----------------------------------------*/
      wxString s_channel_short_name
             = re_extinf.GetMatch( s_line, 3 ).Trim().Trim( false ) ;
      /*-------------------------------------------------------------------*/
      if( s_channel_short_name.empty() )
      {  wxLogError( _( "\"%s\" line %d : Channel name undefined" ),
                     s_origin, i_line_num
                   ) ;
         return( -5 ) ;
      }
      /*--( Extract channel number for the freebox playlist )--------------*/
      int i_channel_num = -1 ;
      if( pl_known == PK_FBTV )
      {  i_channel_num = wxAtoi( re_extinf.GetMatch( s_line, 2 ) ) ;
         if( i_channel_num <= 0 ) { i_channel_num = -1 ; }
      }
      /*-------------------------------------------------------------------*/
      str_stream stream( re_extinf.GetMatch( s_line, 5 ).Trim().Trim( false )
                       ) ;
      /*-------------------------------------------------------------------*/
      wxString s_vlcopt ;

      /*--( Options extraction : they will be added before play )----------*/
      while(    !tis.GetInputStream().Eof()
             && ( s_line = mtf::read_m3u_line( tis, i_line_num )
                ).StartsWith( "#EXTVLCOPT:", &s_vlcopt )
           )
      {
         /*--( Only the track id needs a special treatment )---------------*/
         if( s_vlcopt.BeforeFirst( '=' ) != "audio-track-id" )
         {  stream.m_vs_option.push_back( s_vlcopt ) ; }
         else
         {  /*-------------------------------------------------------------*/
            stream.m_i_audio_track = wxAtoi( s_vlcopt.AfterLast( '=' ) ) ;
            /*-------------------------------------------------------------*/
            if( stream.m_i_audio_track <= 0 )
            {  wxLogError( _( "\"%s\" line %d : Invalid track number" ),
                           s_origin, i_line_num
                         ) ;
               return( -6 ) ;
            }
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      }

      /*--( If the stream URL is not present the entry is simply skipped )-*/
      stream.m_s_url = s_line ;
      if( stream.m_s_url.empty() || stream.m_s_url.StartsWith( '#' ) )
      {  continue ; }

      /*--( Build the associated channel )---------------------------------*/
      CChannel *p_channel_new
      = ( pl_known == PK_NB
          ? new CChannel( s_pl_name, s_channel_short_name, i_channel_num )
          : new CChannel( pl_known, s_channel_short_name, i_channel_num )
        ) ;

      /*--------------------------------------------------------------------+
      ! Search for an already created channel :                             !
      ! - Search by name                                                    !
      !   If found then the stream is added to it                           !
      ! - Search by number (only for freebox TV : because of naming pbs)    !
      !   If found then the stream is added to it                           !
      ! - Not found :                                                       !
      !   The channel is created and the stream added to it                 !
      !                                                                     !
      ! Note that the stream addition is done after search of URL to avoid  !
      ! duplicate                                                           !
      +--------------------------------------------------------------------*/
      CChannel *p_channel = find_channel( p_channel_new->get_name() ) ;
      if( p_channel == nullptr && pl_known == PK_FBTV )
      {  p_channel = find_channel( p_channel_new->get_num() ) ; }

      /*--( Channel already created, this now one is useless )-------------*/
      if( p_channel != nullptr )
      {  delete p_channel_new ; }
      else /*--( An unknown channel has to be kept )-----------------------*/
      {  p_channel = p_channel_new ;
         add_channel( p_channel ) ;
         /*--( Keep list and channel order )-------------------------------*/
         vs_channel.emplace_back( p_channel->get_name() ) ;
         /*----------------------------------------------------------------*/
      }

      /*--( Add stream )---------------------------------------------------*/
      p_channel->add_stream( stream ) ;
      /*-------------------------------------------------------------------*/
      s_line = mtf::read_m3u_line( tis, i_line_num ) ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   return( 0 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CApp::load_channels_from_file( const wxString &s_filename,
                                   const wxString &s_pl_name,
                                   mtf::t_vec_string &vs_channel,
                                   wxString *p_s_label
                                 )
{
   /*--( First check if the file is present )------------------------------*/
   if( !wxFileName::Exists( s_filename ) ) { return( -1 ) ; }
   /*----------------------------------------------------------------------*/
   std::unique_ptr< wxFileInputStream >
                    up_is( new wxFileInputStream( s_filename ) ) ;
   /*----------------------------------------------------------------------*/
   if( up_is.get() == nullptr || !up_is.get()->IsOk() )
   {  wxLogError( _( "Access error to file \"%s\"" ), s_filename ) ;
      return( -2 ) ;
   }
   /*-----------------------------------------------------------------------+
   ! For the playlist tree entry, the name is the file name and the label   !
   ! can be defined in the "m3u" itself by an "#EXTM3U" parameter.          !
   ! For example, for the file "frtv.m3u" the first line can be :           !
   ! #EXTM3U mtf-label="French/TV"                                          !
   +-----------------------------------------------------------------------*/
   return( load_channels( s_filename,
                          s_pl_name.empty()
                          ? wxFileName( s_filename ).GetName() : s_pl_name,
                          up_is.get(), vs_channel, p_s_label
                        )
         ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CApp::load_channels_from_url( const wxString &s_url,
                                  const wxString &s_pl_name,
                                  mtf::t_vec_string &vs_channel,
                                  wxString *p_s_label
                                )
{
   /*----------------------------------------------------------------------*/
   const wxURI co_uri( s_url ) ;
   /*--( First check if the URI is valid )---------------------------------*/
   if( !co_uri.HasScheme() || !co_uri.HasPath() ) { return( -1 ) ; }

   /*--( Avoid loading URL from URL from URL ... )-------------------------*/
   if( m_boo_loading_channels_from_url ) { return( -2 ) ; }
   /*----------------------------------------------------------------------*/
   CBoolInScope bool_scope( m_boo_loading_channels_from_url ) ;

   /*--( Download the playlist )-------------------------------------------*/
   wxString s_data ;
   int i_err = mtf::download( s_url, s_data, nullptr, 2 ) ;
   if( i_err != CDownloadParam::DE_NO_ERROR_DONE ) { return( -3 ) ; }

   /*--( The stream has to be deleted )------------------------------------*/
   std::unique_ptr< wxStringInputStream
                  > up_is( new wxStringInputStream( s_data ) ) ;
   /*----------------------------------------------------------------------*/
   if( up_is.get() == nullptr ) { return( -4 ) ; }
   /*----------------------------------------------------------------------*/
   i_err = load_channels( s_url,
                          s_pl_name.empty() ? co_uri.GetServer() : s_pl_name,
                          up_is.get(), vs_channel
                        ) ;
   /*----------------------------------------------------------------------*/
   return( i_err ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CApp::load_channels_freebox_tv( mtf::t_vec_string &vs_channel )
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
                = "http://mafreebox.freebox.fr/freeboxtv/playlist.m3u" ;
   const wxString co_s_pk_name = get_pk_name( PK_FBTV ) ;
   int i_err ;
   /*--( Ask for a retry if the access to the freebox fails )--------------*/
   while(    ( i_err = load_channels_from_url( co_s_url, co_s_pk_name,
                                               vs_channel
                                             )
             ) != 0
          && wxMessageBox( _( "It doesn't seem possible to access a "
                              "Freebox\n\nTry again ?"
                            ), _( "Freebox access" ),  wxYES_NO
                         ) == wxYES


        )
   {  ; }
   /*----------------------------------------------------------------------*/
   return( i_err ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_image_fn()
{
   /*-----------------------------------------------------------------------+
   ! Loading logo file names across directory tree                          !
   +-----------------------------------------------------------------------*/
   wxArrayString as_logo_full_path ;
   wxDir::GetAllFiles( wxGetApp().m_s_media_path, &as_logo_full_path,
                       "*.png", wxDIR_FILES | wxDIR_DIRS
                     ) ;
   /*--( Conversion to a map to speed up finds )---------------------------*/
   wxString s_name ;
   /*----------------------------------------------------------------------*/
   for( const auto &s_fn : as_logo_full_path )
   {
      /*--( Extract basename and prepare it for a "generous" search )------*/
        s_name
      = s_fn.AfterLast( wxFileName::GetPathSeparator() ).BeforeLast( '.' ) ;
      /*-------------------------------------------------------------------*/
      m_map_image_fn[ mtf::prepare_string( s_name ) ] = s_fn ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_channels()
{
   /*--( Load "en option" channel list )-----------------------------------*/
   wxString s_fn_fbtvo = m_s_media_fb_path + "fbtvo.txt" ;
   mtf::t_map_strint map_fbtvbo ;
   if( wxFileName::Exists( s_fn_fbtvo ) )
   {
      /*-------------------------------------------------------------------*/
      wxLogNull no_log ; // Suppress useless warnings
      wxFileInputStream is_fbtvo( s_fn_fbtvo ) ;
      wxTextInputStream tis_fbtvo( is_fbtvo ) ;
      /*-------------------------------------------------------------------*/
      while( is_fbtvo.IsOk() )
      {  map_fbtvbo[ mtf::prepare_string( tis_fbtvo.ReadLine() ) ] = 1 ; }
      /*-------------------------------------------------------------------*/
   }

   /*--( Filling channels information )------------------------------------*/
   for( auto it : m_map_channel )
   {
      /*-------------------------------------------------------------------*/
      CChannel *p_channel = it.second ;
      /*--( Is this channel part of "en option" freebox channels ? )-------*/
      if( p_channel->get_pl_known() == PK_FBTV )
      {  p_channel->set_fbtvo(   map_fbtvbo.find( p_channel->get_name_find())
                              != map_fbtvbo.end()
                             ) ;
      }
      /*--( To be more "user friendly" the streams are sorted on names )---*/
      p_channel->init_fixed_streams() ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   load_channel_config_info() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::init_playlists()
{
   /*----------------------------------------------------------------------*/
   wxASSERT( m_frame != nullptr ) ;
   /*----------------------------------------------------------------------*/
   CBookPlaylist *bopl = m_frame->m_bopl ;
   wxASSERT( bopl != nullptr ) ;
   wxConfigBase *p_conf = wxConfigBase::Get() ;
   /*----------------------------------------------------------------------*/
   p_conf->SetPath( CPlaylist::m_st_co_p_c_cfg_path ) ;

   /*----------------------------------------------------------------------*/
   mtf::t_vec_string vs_channel ;

   /*----------------------------------------------------------------------*/
   if( M_boo_freebox.get() && load_channels_freebox_tv( vs_channel ) != 0 )
   {  M_boo_freebox.set( false ) ; }

   /*-----------------------------------------------------------------------+
   ! The FB TV flat file is loaded only if the Freebox can't be reached and !
   ! the user requested a forced access to it                               !
   +-----------------------------------------------------------------------*/
   if( !M_boo_freebox.get() && m_boo_forcefreebox )
   {  load_channels_from_file( m_s_media_fb_path + "fbtv.m3u", vs_channel );}

   /*--( Freebox playlists displayed ? )-----------------------------------*/
   if( M_boo_freebox.get() || m_boo_forcefreebox )
   {
      /*-------------------------------------------------------------------*/
      const wxString co_s_fb_label = "Freebox" ;
      /*-------------------------------------------------------------------*/
      bopl->add_pl( CPlaylist::st_new_pl_source( PK_FBTV, co_s_fb_label,
                                                 vs_channel
                                               )
                  ) ;
      bopl->add_pl( CPlaylist::st_new_pl_fbtvb() ) ;
      bopl->add_pl( CPlaylist::st_new_pl_fbtvo() ) ;
      /*--( The TvoOk channel list is stored in the config file )----------*/
      wxString s_channel_list = p_conf->Read( get_pk_name( PK_FBTVO_OK ) ) ;
      bopl->add_pl( CPlaylist::st_new_pl_fbtvo_ok( s_channel_list ) ) ;
      /*-------------------------------------------------------------------*/
      load_channels_from_file( m_s_media_fb_path + "fbradio.m3u",
                               vs_channel
                             ) ;
      bopl->add_pl( CPlaylist::st_new_pl_source( PK_FBRADIO, co_s_fb_label,
                                                 vs_channel
                                               )
                  ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Add the live )----------------------------------------------------*/
   bopl->add_pl( CPlaylist::st_new_pl_live_node() ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_DM, p_conf ) ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_TW, p_conf ) ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_YT, p_conf ) ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_US, p_conf ) ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_PE, p_conf ) ) ;
   bopl->add_pl( CPlaylist::st_new_pl_live( PK_LIVE_MX, p_conf ) ) ;

   /*-----------------------------------------------------------------------+
   ! Loading all "m3u" files across directory tree                          !
   +-----------------------------------------------------------------------*/
   wxArrayString as_m3u_full_path ;
   wxDir::GetAllFiles( m_s_media_path, &as_m3u_full_path,
                       "*.m3u", wxDIR_FILES | wxDIR_DIRS
                     ) ;
   /*--( The files are treated in alphabetical order of path + name )------*/
   as_m3u_full_path.Sort() ;
   wxFileName fn_m3u ;
   wxString s_label ;
   wxFileName fn_label ;
   for( const auto &s_fn_m3u : as_m3u_full_path )
   {

      /*--( Freebox path is a special case treated separately )------------*/
      if( s_fn_m3u.StartsWith( m_s_media_fb_path ) ) { continue ; }
      /*-------------------------------------------------------------------*/
      fn_m3u = s_fn_m3u ;
      fn_m3u.MakeRelativeTo( m_s_media_path ) ;

      /*--( Add TNT Web channels )-----------------------------------------*/
      if( fn_m3u.GetName() == CChannel::m_st_co_p_c_std_french_tv_source )
      {
         /*----------------------------------------------------------------*/
         const wxString s_pl = CChannel::m_st_co_p_c_std_french_tv_source ;
         /*----------------------------------------------------------------*/
         add_channel( CChannelDynamicKnown::CDTK_MYTF1,
                      s_pl, "tf1", "TF1", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_FTV,
                      s_pl, "2", "france 2", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_FTV,
                      s_pl, "3", "france 3", vs_channel
                    ) ;
         add_channel_dailymotion( s_pl, "x5gv6be", "CANAL+", vs_channel ) ;
         add_channel( CChannelDynamicKnown::CDTK_FTV,
                      s_pl, "5", "france 5", vs_channel
                    ) ;
         /*@M6*/
         add_channel_m3u( s_pl, "arte",
                          "https://artelive-lh.akamaihd.net/i"
                          "/artelive_fr@344805/master.m3u8",
                          vs_channel
                        ) ;
         add_channel_dailymotion( s_pl, "x5gv5rr", "C8", vs_channel ) ;
         /*@W9*/
         add_channel( CChannelDynamicKnown::CDTK_MYTF1,
                      s_pl, "TMC", "TMC", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_MYTF1,
                      s_pl, "TFX", "TFX", vs_channel
                    ) ;
         add_channel_m3u( s_pl, "NRJ12",
                          "https://nrj12hls-lh.akamaihd.net/i"
                          "/nrj12hls_1@579113/master.m3u8",
                           vs_channel
                        ) ;
         add_channel_dailymotion( s_pl, "xgepjr",
                      wxString::FromUTF8( u8"La cha\u00eene parlementaire" ),
                                  vs_channel
                                ) ;
         add_channel( CChannelDynamicKnown::CDTK_FTV,
                      s_pl, "4", "france 4", vs_channel
                    ) ;

         add_channel_dailymotion( s_pl, "xgz4t1", "BFM TV", vs_channel ) ;
         /*  add_channel( CChannelDynamicKnown::CDTK_BFM,
                          s_pl, "bfmtv", "BFM TV", vs_channel
                        ) ;
         */

         add_channel_dailymotion( s_pl, "x3b68jn", "CNews", vs_channel ) ;
         add_channel_dailymotion( s_pl, "x5gv5v0", "CStar", vs_channel ) ;
         add_channel( CChannelDynamicKnown::CDTK_GULLI,
                      s_pl, wxEmptyString, "Gulli", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_FTV, s_pl, "O",
                      wxString::FromUTF8( u8"france \u00D4" ), vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_MYTF1,
                      s_pl, "TF1-SERIES-FILMS",
                      wxString::FromUTF8( u8"TF1 S\u00e9ries Films" ),
                      vs_channel
                    ) ;
         add_channel_dailymotion( s_pl, "x2lefik", "L'Equipe", vs_channel ) ;
         /*@6ter*/
         add_channel_dailymotion( s_pl, "x13p3qi", "Numero 23", vs_channel );
         add_channel( CChannelDynamicKnown::CDTK_BFM,
                      s_pl, "rmcdecouverte",
                      wxString::FromUTF8( u8"RMC D\u00e9couverte" ),
                      vs_channel
                    ) ;
         add_channel_m3u( s_pl, wxString::FromUTF8( u8"Ch\u00e9rie 25" ),
                          "https://cherie25hls-lh.akamaihd.net/i"
                          "/cherie25hls_1@579096/master.m3u8",
                          vs_channel
                        ) ;
         add_channel( CChannelDynamicKnown::CDTK_MYTF1,
                      s_pl, "LCI", "LCI", vs_channel
                    ) ;
         /*----------------------------------------------------------------*/
         add_channel( CChannelDynamicKnown::CDTK_BFM,
                      s_pl, "bfmbusiness", "BFM Business", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_BFM,
                      s_pl, "bfmparis", "BFM Paris", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_BFM,
                      s_pl, "rmcsport", "BFM Sport", vs_channel
                    ) ;
         add_channel( CChannelDynamicKnown::CDTK_MELODY,
                      s_pl, wxEmptyString, "Melody", vs_channel
                    ) ;
         /*----------------------------------------------------------------*/
      }

      /*--( The code name of the pl is its system file name )--------------*/
      if( load_channels_from_file( s_fn_m3u, fn_m3u.GetName(), vs_channel,
                                   &s_label
                                 ) == 0
        )
      {  /*-----------------------------------------------------------------+
         ! How is set the path and name/label of the playlist in the tree ? !
         ! - if "mtf-label" is not set then the path and name of the        !
         !   playlist file is used. The path is relative to the "data"      !
         !   directory                                                      !
         ! - if "mtf-label" is set with a RELATIVE path then the file path  !
         !   (relatively to "data") is prepended to it                      !
         ! - if "mtf-label" is set with an ABSOLUTE path then it is this    !
         !   path that is used                                              !
         +-----------------------------------------------------------------*/
         if( s_label.empty() )
         {  fn_label = fn_m3u ; }
         else
         if( s_label.StartsWith( '/' ) )
         {  fn_label = s_label ; }
         else
         {  fn_label = fn_m3u.GetPath() + '/' + s_label ; }
         /*----------------------------------------------------------------*/
         CPlaylist *pl
         = CPlaylist::st_new_pl_source( fn_m3u.GetName(),
                                        fn_label.GetName(),
                                        fn_label.GetPath(), vs_channel
                                      ) ;
         /*----------------------------------------------------------------*/
         bopl->add_pl( pl ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   bopl->add_pl( CPlaylist::st_new_pl_user_node() ) ;
   bopl->add_pl( CPlaylist::st_new_pl_search() ) ;

   /*----------------------------------------------------------------------*/
   {
      /*--------------------------------------------------------------------+
      ! Create user playlists                                               !
      ! The list of the pl user playlists is in the config file. This list  !
      ! is ordered.                                                         !
      ! The channel list of each pl is in the config file too.              !
      +--------------------------------------------------------------------*/
      wxStringTokenizer tkz_pl_user( M_s_bopl_pl_user_list.get(), '\n' ) ;
      wxString s_pl_name ;
      wxString s_channel_list ;
      /*-------------------------------------------------------------------*/
      while( tkz_pl_user.HasMoreTokens() )
      {
         /*--( Based on pl name the channel list is extracted from config )*/
         s_pl_name = tkz_pl_user.GetNextToken() ;
         p_conf->Read( s_pl_name, &s_channel_list, "" ) ;
         /*----------------------------------------------------------------*/
         bopl->add_pl( CPlaylist::st_new_pl_user( s_pl_name, s_channel_list )
                     ) ;
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*--( Initialise playlists constant elements )--------------------------*/
   {
      /*--( We need one playlist to make the changes )---------------------*/
      CPlaylist *pl = get_pk_pl( PK_USER_NODE ) ;

      /*--( First set the different fonts )--------------------------------*/
      const wxFont font_normal = pl->GetFont().Larger() ;
      const wxFont font_played_active = pl->GetFont().Scaled( 2 ).Bold() ;
      const wxFont font_played_other  = pl->GetFont().Scaled( 2 ).Italic() ;
      /*--( Cf "wxListCtrlBase::EnableAlternateRowColours" wxWidgets code )*/
      const wxColor color_bg = pl->GetBackgroundColour() ;
      int i_alpha = ( color_bg.GetRGB() > 0x808080 ? 97 : 150 ) ;
      wxColor color_bg_alt = color_bg.ChangeLightness( i_alpha ) ;
      /*-------------------------------------------------------------------*/
      m_tb_lia[ 0 ].SetFont( font_normal ) ;
      m_tb_lia[ 1 ].SetFont( font_normal ) ;
      m_tb_lia[ 1 ].SetBackgroundColour( color_bg_alt ) ;
      /*-------------------------------------------------------------------*/
      m_tb_lia_active[ 0 ].SetFont( font_played_active ) ;
      m_tb_lia_active[ 1 ].SetFont( font_played_active ) ;
      m_tb_lia_active[ 1 ].SetBackgroundColour( color_bg_alt ) ;
      /*-------------------------------------------------------------------*/
      m_tb_lia_other[ 0 ].SetFont( font_played_other ) ;
      m_tb_lia_other[ 1 ].SetFont( font_played_other ) ;
      m_tb_lia_other[ 1 ].SetBackgroundColour( color_bg_alt ) ;

      /*--( Then compute the maximum width of the playlist column )--------*/
      int i_width_max = 350 ;
      /*--( Compute "maximum" text width : biggest font used )-------------*/
      wxFont font_sav = pl->GetFont() ;
      pl->SetFont( m_tb_lia_active[ 0 ].GetFont() ) ;
      /*-------------------------------------------------------------------*/
      for( auto it : m_map_channel )
      {
         /*----------------------------------------------------------------*/
           int i_width
         = pl->GetTextExtent( it.second->get_pl_text_long() ).GetWidth() ;
         /*----------------------------------------------------------------*/
         if( i_width > i_width_max ) { i_width_max = i_width ; }
         /*----------------------------------------------------------------*/
      }
      /*--( Add image width and some separation space )--------------------*/
      i_width_max += m_il_logo->GetSize().GetWidth() + 20 ;
      /*-------------------------------------------------------------------*/
      pl->SetFont( font_sav ) ;
      /*--( Keep value for the new playlists )-----------------------------*/
      CPlaylist::st_set_col_width( i_width_max ) ;
      /*--( Now set the column width )-------------------------------------*/
      m_frame->m_bopl->exec_on_tree_pl(
                []( CPlaylist *pl )
                {  pl->SetColumnWidth( 0, CPlaylist::st_get_col_width() ) ; }
                                      ) ;

      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::load_channel_config_info()
{
   /*----------------------------------------------------------------------*/
   wxConfigBase *p_conf = wxConfigBase::Get() ;

   /*--( Path created if not present )-------------------------------------*/
   p_conf->SetPath( CChannel::m_st_co_p_c_cfg_path ) ;
   /*----------------------------------------------------------------------*/
   long     l_index ;
   wxString s_entry ;
   wxString s_val   ;
   /*----------------------------------------------------------------------*/
   if( !p_conf->GetFirstEntry( s_entry, l_index ) )
   {  return ; }
   /*----------------------------------------------------------------------*/
   do
   {  /*--( The entry is the channel name )--------------------------------*/
      CChannel *p_channel = get_channel( s_entry ) ;
      if( p_channel == nullptr ) { continue ; }
      /*-------------------------------------------------------------------*/
      if( p_conf->Read( s_entry, &s_val ) )
      {  p_channel->string_to_datas( s_val ) ; }
      /*-------------------------------------------------------------------*/
   } while( p_conf->GetNextEntry( s_entry, l_index ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CApp::save_channel_config_info()
{
   /*----------------------------------------------------------------------*/
   wxConfigBase *p_conf = wxConfigBase::Get() ;

   /*--( Path created if not present )-------------------------------------*/
   p_conf->SetPath( CChannel::m_st_co_p_c_cfg_path ) ;
   /*----------------------------------------------------------------------*/
   for( auto it : m_map_channel )
   {
      /*-------------------------------------------------------------------*/
      CChannel *p_channel = it.second ;
      /*-------------------------------------------------------------------*/
      if( p_channel->get_save_cfg() )
      {  p_conf->Write( p_channel->get_name(),
                        p_channel->datas_to_string()
                      ) ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--( Save playlist content and tab order )--------------------------------*/
void CApp::save_playlist_config_info()
{
   /*--( Path created if not present )-------------------------------------*/
   wxConfigBase::Get()->SetPath( CPlaylist::m_st_co_p_c_cfg_path ) ;
   /*----------------------------------------------------------------------*/
   m_frame->m_bopl->exec_on_tree_pl(
                      []( CPlaylist *pl )
                      {  if( pl->get_save_cfg() )
                         {  wxConfigBase::Get()->Write( pl->get_name(),
                                                        pl->datas_to_string()
                                                      ) ;
                         }
                      }
                                   ) ;
   /*----------------------------------------------------------------------*/
}

/*--( Find date for the next update check )--------------------------------*/
wxString CApp::next_chkupd( int i_chkupd_freq ) const
{
   /*----------------------------------------------------------------------*/
   if( i_chkupd_freq <= CUF_NEVER || i_chkupd_freq >= CUF_NB )
   {  return( wxEmptyString ) ; }

   /*----------------------------------------------------------------------*/
   wxDateTime dt_next = wxDateTime::Today() ;
   /*----------------------------------------------------------------------*/
   dt_next.ParseISODate( M_s_chkupd_last.get() ) ;
   /*----------------------------------------------------------------------*/
   switch( i_chkupd_freq )
   {
      /*-------------------------------------------------------------------*/
      case CUF_DAY   : dt_next.Add( wxDateSpan( 0, 0, 0, 1 ) ) ; break ;
      case CUF_WEEK  : dt_next.Add( wxDateSpan( 0, 0, 1, 0 ) ) ; break ;
      case CUF_MONTH : dt_next.Add( wxDateSpan( 0, 1, 0, 0 ) ) ; break ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return( dt_next.FormatISODate() ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
e_pl_known CApp::find_pl_known( const wxString &s_name ) const
{
   /*----------------------------------------------------------------------*/
   for( int i_num = 0 ; i_num < PK_NB ; ++i_num )
   {  if( s_name == m_tb_pk_info[ i_num ].m_s_name )
      {  return( static_cast< e_pl_known >( i_num ) ) ; }
   }
   /*----------------------------------------------------------------------*/
   return( PK_NB ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
bool CApp::is_pk_name_reserved( const wxString &s_name ) const
{
   /*----------------------------------------------------------------------*/
   for( const auto &pki : m_tb_pk_info )
   {  if( s_name.CmpNoCase( pki.m_s_name ) == 0 ) { return( true ) ; } }
   return( false ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                           End of file CApp.cpp                            !
+==========================================================================*/
